(** Time-stamp: <modified the 18/12/2007 (at 14:55) by Erwan Jahier> *)


open Printf 
open Lxm
open Errors
open SyntaxTree
open CompiledData

(*----------------------------------------------------
EvalArray_error :
	- leve par les fonctions ddies aux tableaux
----------------------------------------------------*)
exception EvalArray_error of string

(*----------------------------------------------------
EvalConst_error :
	- leve localement dans les sous-fonctions,
	- capte dans eval_const et tranforme en Compile_error.
----------------------------------------------------*)
exception EvalConst_error of string

(*----------------------------------------------------
Utilitaires : erreurs classiques
----------------------------------------------------*)
let type_error (v : const_eff list) (expect : string) =
	raise (EvalConst_error(sprintf "type combination error, %s expected" expect))

(* Utile : erreur d'arit *)
let arity_error (v : const_eff list) (expect : string) =
	raise (EvalConst_error(
		sprintf "arity error, %s instead of %d" expect (List.length v)))

(* Utile : erreur, oprateur interdit  *)
let uneval_error (v : const_eff list) (opname : string) =
	raise (EvalConst_error(
		sprintf "operation %s not allowed in static expression" opname))

(*----------------------------------------------------
	Calcul sur les const_eff
------------------------------------------------------
op_computer :
- associe  chaque oprateur valuable (posop : by_pos_op)
  une fonction de calcul Some (const_eff list -> const_eff_list)
  ou Erreur
----------------------------------------------------*)

let op_computer (posop : by_pos_op) (src: Lxm.t)
    = (* retour : (const_eff list -> const_eff_list) option *)
  (
    (*----------------------------
      le template pour tous les :
      bool*bool->bool
      ----------------------------*)
    let generic_bin_bool
	(nm : string) 
	(f : bool -> bool -> bool)
	(args : const_eff list)
	= (
	  match args with
	      [Bool_const_eff v0; Bool_const_eff v1] -> [Bool_const_eff (f v0 v1)]
	    |	[x0; x1] -> (type_error [x0; x1] "bool*bool")
	    |	x        -> (arity_error x "2" )
	) in
      (*----------------------------
	le template pour tous les :
	int*int->int
	----------------------------*)
    let generic_bin_int
	(nm : string) 
	(f : int -> int -> int)
	(args : const_eff list)
	= (
	  match args with
	      [Int_const_eff v0; Int_const_eff v1] -> [Int_const_eff (f v0 v1)]
	    |	[x0; x1] -> (type_error [x0; x1] "int*int")
	    |	x        -> (arity_error x "2" )
	) in
      (*----------------------------
	le template pour tous les :
	num*num->bool
	N.B. on est oblig de passer 
	2 "copie" du comparateur
	(fi pour int, fr pour float)
	sinon caml ne peut pas typer ...
	----------------------------*)
    let generic_num_comp
	(nm : string)
	(fi : int -> int -> bool)
	(fr : float -> float -> bool)
	(args : const_eff list)
	= (
	  match args with
	      [Int_const_eff v0; Int_const_eff v1] -> (
		[Bool_const_eff (fi v0 v1)]
	      ) |
		  [Real_const_eff v0; Real_const_eff v1] -> (
		    let res = (fr v0 v1) in
		      warning src 
			(sprintf 
			   "float in static exp: %f%s%f evaluated as %b" v0 nm v1 res);
		      [Bool_const_eff res]
		  ) |
		      [x0; x1] -> (type_error [x0; x1] "int*int or real*real")
	    |	x     -> (arity_error x "2" )
	) in
      (*----------------------------
	le template pour tous les :
	num*num->num
	----------------------------*)
    let generic_bin_num
	(nm : string)
	(fi : int -> int -> int)
	(fr : float -> float -> float)
	(args : const_eff list)
	= (
	  match args with
	      [Int_const_eff v0; Int_const_eff v1] -> (
		[Int_const_eff (fi v0 v1)]
	      ) |
		  [Real_const_eff v0; Real_const_eff v1] -> (
		    let res = (fr v0 v1) in
		      warning src 
			(sprintf 
			   "float in static exp: %f%s%f evaluated as %f" v0 nm v1 res);
		      [Real_const_eff res]
		  ) |
		      [x0; x1] -> (type_error [x0; x1] "int*int or real*real")
	    |	x     -> (arity_error x "2" )
	) in
      (*----------------------------
	Calcul du if
	----------------------------*)
    let compute_if (args : const_eff list) = (
      match args with 
	  [Bool_const_eff c; x1; x2] -> (
	    let t1 = type_of_const_eff x1 in	
	    let t2 = type_of_const_eff x2 in	
	      if (t1 = t2 ) then (if (c) then [x1] else [x2])
	      else (type_error args "bool*t*t for some type t")
	  )
	| [x0; x1; x2] -> (type_error args "bool*t*t for some type t")
	| x -> (arity_error x "3")
    ) in
      (*----------------------------
	Calcul de l'galit
	N.B. Sur les constantes abstraites
	on est trs mfiant
	N.B. Sur les types structure,
	on fait des appels rcursifs
	----------------------------*)
    let rec compute_eq 
	(args : const_eff list)
	= (
	  let rec fields_eq f0 f1 = (
	    match (f0, f1) with
		([], []) -> (
		  [Bool_const_eff true]
		) |
		    ((f0,h0)::t0, (f1,h1)::t1) -> (
		      assert (f0 = f1);
		      match (compute_eq [h0;h1]) with
			  [Bool_const_eff false] -> [Bool_const_eff false]
			| [Bool_const_eff true] -> (fields_eq t0 t1) 
			| _ -> assert false
		    )
	      |
		  _ -> assert false
	  ) 
	  in
	    match args with
		[Bool_const_eff v0; Bool_const_eff v1] -> [Bool_const_eff (v0 = v1)]
	      |	[Int_const_eff v0; Int_const_eff v1] -> [Bool_const_eff (v0 = v1)]
	      |	[Real_const_eff v0; Real_const_eff v1] -> (
		  let res = (v0 = v1) in
		    warning src 
		      (sprintf "float in static exp: %f=%f evaluated as %b" v0 v1 res);
		    [Bool_const_eff res]
		)
		  (*
		    2007-07 obsolete

		    |  [Extern_const_eff (v0, t0); Extern_const_eff (v1, t1)] -> (
		    if (t0 <> t1) then (
		    type_error args "t*t for some type t"
		    ) else if (v0 <> v1) then (
		    uneval_error args (
		    sprintf "%s=%s (external constants)"
		    (string_of_fullid v0)
		    (string_of_fullid v1)
		    )
		    ) else (
		    [Bool_const_eff true]
		    )
		    )
		  *)
	      |	[Enum_const_eff (v0, t0); Enum_const_eff (v1, t1)] -> (
		  if (t0 = t1) then [Bool_const_eff (v0 =  v1)]
		  else type_error args "t*t for some type t"
		)
	      |	[Struct_const_eff (f0, t0); Struct_const_eff (f1, t1)] -> (
		  if (t0 = t1) then (fields_eq f0 f1)
		  else type_error args "t*t for some type t"
		)
	      |	[x;y] -> type_error args "t*t for some type t"
	      |	x -> arity_error args "2"
	) in
      (* match principal *)
      match posop with 
	  TRUE_n     -> (
	    function
		[] -> [ Bool_const_eff true ] 
	      | x  -> arity_error x "0"
	  )
	| FALSE_n    -> (
	    function
		[] -> [ Bool_const_eff false ]
	      | x  -> arity_error x "0"
	  )
	| ICONST_n s -> (
	    function  
		[] -> [ Int_const_eff (int_of_string (Ident.to_string s)) ]
	      | x  -> arity_error x "0"
	  )
	| RCONST_n s -> (
	    function
		[] -> [ Real_const_eff (float_of_string (Ident.to_string s)) ]
	      | x  -> arity_error x "0"
	  )
	| IDENT_n  s -> assert false
	    (* unaires *)
	| NOT_n       -> (
	    function
		[Bool_const_eff v] -> [Bool_const_eff (not v)]
	      | [x] -> type_error [x] "bool"
	      |  x  -> arity_error x "1"
	  )
	| UMINUS_n    -> (
	    function
		[Int_const_eff v] -> [Int_const_eff (- v)]
	      | [Real_const_eff v] -> [Real_const_eff (-. v)]
	      | [x] -> type_error [x] "int or real"
	      |  x  -> arity_error x "1"
	  )
	| PRE_n       -> ( function x -> uneval_error x "pre")
	| CURRENT_n   -> ( function x -> uneval_error x "current")
	| REAL2INT_n  -> (
	    function
		[Real_const_eff v] -> (
		  let res = (int_of_float v) in
		    warning src
		      (sprintf "float2int cast: '%f' converted to '%d'" v res);
		    [Int_const_eff res]
		)
	      | [x] -> type_error [x] "real"
	      | x   -> arity_error x "1"
	  )
	| INT2REAL_n  -> (
	    function
		[Int_const_eff v] -> [Real_const_eff (float_of_int v)]
	      | [x] -> type_error [x] "int"
	      | x   -> arity_error x "1"
	  )
	    (* binaires *)
	| ARROW_n  -> ( function x -> uneval_error x "->")
	| WHEN_n   -> (
	    function x -> uneval_error x "when"
	  )
	    (* pour les operateurs binaire boolens, on a un seul "template" *)
	| AND_n    -> ( generic_bin_bool "and" (&&) )
	| OR_n     -> ( generic_bin_bool "or" (or) )
	| XOR_n    -> ( generic_bin_bool "xor" (<>) )
	| IMPL_n   -> ( generic_bin_bool "=>" (function x -> function y -> not x or y))
	| EQ_n     -> ( compute_eq )
	| NEQ_n    -> (
	    function x -> (
	      match (compute_eq x) with
		  [Bool_const_eff v] -> [Bool_const_eff (not v)]
		|	x -> x 
	    )
	  )
	| LT_n     -> ( generic_num_comp "<" (<) (<) )
	| LTE_n    -> ( generic_num_comp "<=" (<=) (<=) )
	| GT_n     -> ( generic_num_comp ">" (>) (>) )
	| GTE_n    -> ( generic_num_comp ">=" (>=) (>=) )
	| DIV_n    -> ( generic_bin_int  "div" ( / ) )
	| MOD_n    -> ( generic_bin_int  "mod" ( mod ) )
	| MINUS_n  -> ( generic_bin_num "-" (-) (-.) )
	| PLUS_n   -> ( generic_bin_num "+" (+) (+.) )
	| SLASH_n  -> ( generic_bin_num "/" (/) (/.) )
	| TIMES_n  -> ( generic_bin_num "*" ( * ) ( *. ) )
	| POWER_n  -> assert false
	    (* ternaires *)
	| IF_n   -> (compute_if)
	    (* n-aires *)
	| NOR_n     -> assert false
	| DIESE_n   -> assert false
	| CALL_n s  -> assert false
	    (* pseudo-unaire : acces structure *)
	| STRUCT_ACCESS_n fid -> (
	    function 
		[Struct_const_eff (flst, typ)] -> (
		  try [(List.assoc (fid) flst)]
		  with Not_found -> 
		    raise (
		      EvalConst_error
			     (sprintf 
				"%s is not a field of struct %s" 
				(Ident.to_string fid) (string_of_type_eff(typ))
			     ))
		)
	      | [x] -> type_error [x] "struct type"
	      | x -> arity_error x "1"
	  ) |
	      _ -> assert false
  )

(*----------------------------------------------------
Utilitaire :
extraire une tranche de tableau

N.B. first_ix  last_ix step et width sont supposs 
venir de eva et donc tre corrects

N.B. Puisque correct, last_ix est inutile, mais bon ...
-----------------------------------------------------*)
let make_slice_const
	(ctab : const_eff array)
	(ctype : type_eff)
	(slice : slice_eff)
= (
	let get_res (ix : int) = (
		Array.get ctab (slice.se_first + ix*slice.se_step) 
	) in
	[Array_const_eff (Array.init slice.se_width get_res, ctype)]
)

(*----------------------------------------------------
Utilitaire : fabriquer si possible une const tableau
-----------------------------------------------------*)
let make_array_const (ops : const_eff list array) = (
  let expected_type = ref None in
  let treat_arg (op : const_eff list) = (
    match op with
	(* non tuple *)
	[x] -> (
	  let xtyp = type_of_const_eff x in
	    match (!expected_type) with
		None -> (
		  expected_type := Some xtyp ;
		  x
		) |
		    Some t -> (
		      if(t = xtyp) then x
		      else 
			raise (
			  EvalConst_error(
			    "type error in array, "^
			      (string_of_type_eff xtyp)^
			      " mixed with " ^string_of_type_eff t
			  ))
		    )
	) 
      | _  ->  
	  (* tuple *)
	  raise (EvalConst_error("array of tuple not allowed"))
  ) in 
    (* corps de make_array_const *)
  let res = Array.map treat_arg ops in
    match (!expected_type) with
	(* tableau null *)
	None -> raise (EvalConst_error("empty array"))
      |	Some t -> Array_const_eff(res, t)
)
(*----------------------------------------------------
Utilitaire : fabriquer si possible une cons structure 
N.B. Par construction on sait que nops n'a pas de doublons
-----------------------------------------------------*)
let make_struct_const 
    (teff : type_eff)
    (arg_tab : (Ident.t, Lxm.t * const_eff) Hashtbl.t) =
  (
    (* on verifie qu'on a bien un type struct *)
    match teff with
	Struct_type_eff (tnm, flst) -> (
	  (* on construit la liste dans le BON ordre *)
	  let make_eff_field ((fn: Ident.t),(ft:type_eff),(fv:const_eff option)) = (
	    try (
	      (* on prend en priorit dans arg_tab *)
	      match (Hashtbl.find arg_tab fn) with
		  (lxm, v) -> (
		    (* effet de bord : on vire la valeur de arg_tab *)
		    Hashtbl.remove arg_tab fn ;
		    let vt = type_of_const_eff v in
		      if (vt = ft) then (fn, v) (*ok*)
		      else raise (Compile_error(
				    lxm , 
				    sprintf
				      "type error in struct %s, %s instead of %s"
				      (Ident.string_of_long tnm)
				      (string_of_type_eff vt)
				      (string_of_type_eff ft) 
				  )) 
		  )
	    ) with Not_found -> (
	      (* sinon la valeur par dfaut *)
	      match fv with
		  Some v -> (fn, v) (* ok : v correcte par construction *)
		| None -> 
		    raise (EvalConst_error(
			     sprintf
			       "bad struct expression, no value given for field %s" 
					     (Ident.to_string fn)
			   ))
	    )
	  ) in
	    (* on mappe flst pour avoir la liste dans le bon ordre *)
	  let eff_fields = List.map make_eff_field flst in
	    (* si arg_tab n'est pas vide, erreur sur le premier *) 
	  let raise_error (id : Ident.t) ((lxm : Lxm.t), (veff : const_eff)) 
	      = raise(Compile_error(
			lxm, 
			sprintf
			  "%s is not a field of struct %s" 
			  (Ident.to_string id) (string_of_type_eff(teff))
		      )) 
	  in 
	    Hashtbl.iter raise_error arg_tab ;
	    (* ok : tout s'est bien pass ! *)
	    Struct_const_eff (eff_fields, teff)
	) |
	    _ -> raise (EvalConst_error(
			  sprintf
			    "struct type expected instead of %s" 
			    (string_of_type_eff teff)
			))
  )

(************DEBUT SCORIE***************************************

(*----------------------------------------------------
SCORIES : PAS D'EXTENTION HOMOMORPHE IMPLICITE EN V6
	Application d'une opration homomorphe classique
------------------------------------------------------
Si TOUS les args sont des tableaux de mme taille,
alors on en conclue qu'il s'agit d'une extension
homomorphe
----------------------------------------------------*)

(*
Transforme (si possible)
une liste de tableaux en tableau de liste.
exemple (c'est plus parlant !):
get_extension_args [
	Array_const_eff([| a1; a2; ...; an |],ta) ;
	Array_const_eff([| b1; b2; ...; bn |],tb) ;
	...
	Array_const_eff([| z1; z2; ...; zn |],zb) ;
] =
Some [ [a1; b1; ...; z1]; ... [an; bn; ...; zn] ]
Dans les autres cas => none
*)

let get_extension_args (clist : const_eff list) = (
	let treat_arg (c : const_eff) (acc : const_eff list array option) = (
		match c with
		Array_const_eff (ctab, typ) -> (
			match acc with
			None -> Some (Array.map (function x -> [x]) ctab)
			| Some res -> Some (Utils.array_map2 (fun x -> fun y -> x::y) ctab res)
		) | _ -> raise (Invalid_argument "")
	)	
	in try
		List.fold_right treat_arg clist None 
	with _ -> None 
)

let rec compute_homomorphic_op 
	(oper : const_eff list -> const_eff list)
	(args : const_eff list) =
(
	match (get_extension_args args) with
	None -> (
		(* pas extension => on opere juste *)
		oper args
	) |
	Some c_lst_tab -> (
		(* extension *) 
		let elts = Array.map (compute_homomorphic_op oper) c_lst_tab in
		[ make_array_const elts ]
	)
)
***********FIN SCORIE***************************************)

(*----------------------------------------------------
	Evaluation rcursive des expressions constantes
------------------------------------------------------
eval_const :
	- entres :  id_solver et val_exp
	- sortie :        const_eff list
	- Effet de bord : Compile_error 
Rle :
	-> rsoud les rfrences aux idents
	-> gre les appels rcursifs (valuation des arguments) 
----------------------------------------------------*)
let rec eval_const
    (env : id_solver) 
    (vexp : val_exp)
    = (
      (*-----------------------------------
	fonction rcursive principale
	-> capte les nv
	-> rcupre les EvalConst_error 
	-----------------------------------*)
      let rec rec_eval_const (vexp : val_exp) = (
	match vexp with
	    CallByPos  ({it=posop; src=lxm}, args) -> (
	      try eval_by_pos_const posop lxm args
	      with EvalConst_error msg ->
		raise (Compile_error(lxm, "can't eval constant: "^msg))
	    ) 
	  |
	    CallByName ({it=nmop; src=lxm}, nmargs ) -> (
	      try eval_by_name_const nmop lxm nmargs
	      with EvalConst_error msg ->
		raise (Compile_error(lxm, "can't eval constant: "^msg))
	    )
      )
	(*-----------------------------------
	  fonction rcursive secondaire
	  eval. exp classique (by pos)
	  N.B. On distingue les oprations 
	  classiques (avec extention tableau
	  implicie) des autres. Ici, on traite
	  toutes les oprations non classiques. 
	  -----------------------------------*)
      and eval_by_pos_const
	  (posop : by_pos_op)   (* l'operateur *)
	  (lxm : Lxm.t)      (* source de l'oprateur *)
	  (args : val_exp list) (* arguments *)
	  = (
	    match (posop) with 
		(* capte les idents de constantes *)
		IDENT_n  id  -> (
		  (* 2007-07 on interdit les externes *)
		  match (env.id2const id lxm) with
		      Extern_const_eff _ -> (
			raise (EvalConst_error( sprintf
						  "external constant not allowed (%s)"
						  (Lxm.details lxm)))
		      ) |
			  x ->  [ x ]
		)
		  (* oprateur lazzy *)
	      | WITH_n -> (
		  match args with
		      [a0;a1;a2] -> (
			match (rec_eval_const a0) with
			    [ Bool_const_eff true] -> rec_eval_const a1
			  | [ Bool_const_eff false] -> rec_eval_const a2
			  | x -> type_error x "bool"
		      )  |
			  _ -> raise (EvalConst_error( 
					sprintf
					  "arity error: 3 extected instead of %d" 
					  (List.length args)))
		)
		  (* mettre  plat la liste des args *)
	      | TUPLE_n -> ( List.flatten (List.map rec_eval_const args))
		  (* les tableaux de tuples sont interdits *)
	      | HAT_n -> (
		  match args with
		      [cexp; szexp] -> (
			try (
			  let sz = eval_array_size env szexp in
			    match rec_eval_const cexp with
				[cst] -> (
				  let atab = Array.make sz cst in	
				    [ Array_const_eff (atab, type_of_const_eff cst) ]
				) 
			      |
				x -> 
				 raise (EvalConst_error(
					       "array of tuple not allowed"))
			) with 
			    EvalArray_error msg -> raise(EvalConst_error msg)
		      ) 
		    |
		      _ -> raise(EvalConst_error
				   (sprintf
				      "arity error: 2 exptected instead of %d" 
				      (List.length args)
				   )
				)
		)
	      | CONCAT_n -> (
		  let ops = (List.map rec_eval_const args) in
		    match ops with
			[[c0];[c1]] -> (
			  match (c0, c1) with
			      (Array_const_eff (v0, t0), 
			       Array_const_eff (v1, t1) ) -> (
				if(t0 = t1) then 
				  [Array_const_eff (Array.append v0 v1, t0)]	
				else 
				  raise(EvalConst_error(
					  sprintf
					    "type combination error, can't concat %s with %s"
					    (string_of_type_eff(type_of_const_eff c0))
					    (string_of_type_eff(type_of_const_eff c1)) 
							  ))
			      ) 
			    | _ -> raise(
				EvalConst_error(
				  "type combination error, array type expected"))
			) 
		      | _ -> raise(EvalConst_error
				     (sprintf
					"arity error: 2 extected instead of %d" 
					(List.length ops)))
		)
	      | ARRAY_n -> (
		  let ops = (List.map rec_eval_const args) in
		    [make_array_const (Array.of_list ops)]
		)
	      | ARRAY_ACCES_n ix -> (
		  let effargs = List.flatten (List.map rec_eval_const args) in
		    match effargs with
			[x] -> (
			  match x with
			      Array_const_eff (elts, typelts) -> (
				try (
				  let sz = Array.length elts in
				  let effix = eval_array_index env ix sz in
				    [Array.get elts effix ]
				) 
				with EvalArray_error msg -> raise(EvalConst_error msg)
			      ) |
				  _ -> type_error effargs "some array" 
				    
			) |
			    _ -> type_error effargs "some array"
		)
	      | ARRAY_SLICE_n sl -> (
		  (* value l'argument *)
		  let (elts, typelts)  = (
		    match List.flatten (List.map rec_eval_const args) with
			[Array_const_eff (l, t)] -> (l, t) 
		      | x -> type_error x "some array"
		  ) in 
		    (* on en dduit la taille du tableau *) 
		  let sz = Array.length elts in
		    (* value la slice *)
		    try (
		      let sliceff = eval_array_slice env sl sz lxm in
			make_slice_const elts typelts sliceff
		    ) with
			EvalArray_error msg -> raise(EvalConst_error msg)
		) |
		    (* extension homomorphe spciale ? *)
		    CALL_n s -> (
		      print_string "*** node call not yet implemented\n";
		      assert false
		    ) 
	      | 
		  (* PAS D'EXTENTION HOMOMORPHE IMPLICITE
		     
		     Toutes les autres oprations sont des 
		     cas classiques d'extention homomorphe *)
		  _ -> 
		  (
		    let effargs = List.flatten (List.map rec_eval_const args) in
		      (* compute_homomorphic_op (op_computer posop lxm) effargs *)
		      (op_computer posop lxm) effargs
		  )
	  ) (* FIN DE : eval_by_pos_const *)
	(*-------------------------------------*)
	(* Fonction rcursive secondaire       *)
	(*-------------------------------------*)
	(* -> Eval. d'une expression spciale  *)
	(* "par nom"                           *)
	(*-------------------------------------*)
      and eval_by_name_const
	  (namop : by_name_op)   (* l'operateur *)
	  (lxm : Lxm.t)      (* source de l'oprateur *)
	  (namargs : (Ident.t srcflaged * val_exp) list) (* arguments *)
	  = (
	    match namop with
	      |	STRUCT_EXP_n -> assert false
		  
	      | BYNAME_CALL_n opid -> (
		  (* effet de bord : on tabule les param effectif *)
		  let arg_tab = Hashtbl.create 50 in 
		  let treat_one_arg ((pid:Ident.t srcflaged), (pexp:val_exp)) = (
		    if (Hashtbl.mem arg_tab pid.it) then (
		      raise(EvalConst_error(
			      sprintf
				"multiple definition of param %s in %s call"
				(Ident.to_string pid.it)
				(Ident.string_of_idref opid) 
			    ))	
		    ) else (
		      let v = rec_eval_const pexp in
			match v with
			    [x] -> Hashtbl.add arg_tab pid.it (pid.src, x)
			  | _ -> 
			      raise(
				EvalConst_error(
				  sprintf
				    "unexpected tuple value for param %s in %s call"
				    (Ident.to_string pid.it)
				    (Ident.string_of_idref opid)
				))
		    )
		  ) in List.iter treat_one_arg namargs ;
		    (* pour l'instant, on ne traite que les constructions de struct *)
		    try ( 
		      let teff = env.id2type opid lxm in
			[make_struct_const teff arg_tab]
		    ) with 
			_ -> raise(EvalConst_error(
				     sprintf
				       "struct type expected instead of %s"
				       (Ident.string_of_idref opid)
				   ))
		)
	  ) (* FIN DE : eval_by_name_const *)
	(*-------------------------------------*)
	(* Corps de la fonction principale      *)
	(*-------------------------------------*)
      in rec_eval_const vexp 
    ) (* fin de eval_const *)
  (*---------------------------------------------------------------------
    eval_array_size
    -----------------------------------------------------------------------
    Rle : calcule une taille de tableau 

    Entres: 

    Sorties :
    int (strictement positif)

    Effets de bord :
    EvalArray_error "bad array size, type int expected but get <t>" si t pas int
    EvalArray_error "bad array size <n>" si n <= 0
    ----------------------------------------------------------------------*)
and eval_array_size
    (env : id_solver)
    (szexp : val_exp)
    = (
      match (eval_const env szexp) with
	| [Int_const_eff sz] -> (
	    if (sz > 0) then sz
	    else raise(EvalArray_error(sprintf "bad array size %d" sz))
	  ) 
	| [x] -> raise(EvalArray_error(
			 sprintf 
			   "bad array size, int expected but get %s"
			   (string_of_type_eff(type_of_const_eff x)))
		      ) 
	| _ -> raise(EvalArray_error(
		       sprintf 
			 "bad array size, int expected but get a tuple")
		    )
    )
  (*---------------------------------------------------------------------
    eval_array_index
    -----------------------------------------------------------------------
    Rle :

    Entres :
    id_solver, val_exp, taille du tableau

    Sorties :
    int (entre 0 et taille du tableau -1

    Effets de bord :
    EvalArray_error msg si pas bon
    ----------------------------------------------------------------------*)
and eval_array_index
    (env : id_solver)
    (ixexp : val_exp)
    (sz : int)
    = (
      match (eval_const env ixexp) with
	  [Int_const_eff i] -> (
	    if ((i >= 0) && (i < sz)) then i
	    else raise(EvalArray_error(
			 sprintf
			   "array index %d out of bounds 0..%d" i (sz-1)))
	  ) 
	| [x] -> raise(EvalArray_error(sprintf 
					 "bad array index, int expected but get %s"
					 (string_of_type_eff(type_of_const_eff x)))
		      ) 
	| _ -> raise(EvalArray_error(sprintf 
				       "bad array index, int expected but get a tuple")
		    )

    )
  (*---------------------------------------------------------------------
    eval_array_slice
    -----------------------------------------------------------------------
    Rle :

    Entres :
    id_solver, slice_info, size du tableau,
    lxm (source de l'opration slice pour warning)
    Sorties :
    slice_eff, i.e.
    (fisrt,last,step,width) tels que step <> 0 et
    - si step > 0 alors 0<=first<=last et first<=sz
    - si step < 0 alors 0<=last<=first et first<=sz
    - 1<=width<=sz 
    Effets de bord :
    EvalArray_error msg si pas bon
    ----------------------------------------------------------------------*)
and eval_array_slice
    (env : id_solver)
    (sl : slice_info)
    (sz : int)
    (lxm : Lxm.t)
    = (
      let first_ix = eval_array_index env sl.si_first sz in 
      let last_ix = eval_array_index env sl.si_last sz in
      let step = (
	match sl.si_step with
	  | Some stepexp -> (
	      match (eval_const env stepexp) with
		| [Int_const_eff s] ->  s (* ok *)
		    
		| [x] -> raise(EvalArray_error(
				 sprintf 
				   "bad array step, int expected but get %s"
				   (string_of_type_eff(type_of_const_eff x)))
			      ) 
		| _ -> raise(EvalArray_error(
			       sprintf 
				 "bad array step, int expected but get a tuple")
			    )
	    ) 
	  | None -> if (first_ix <= last_ix) then 1 else -1
      ) 
      in
	if (
	  (step = 0)
	  || (step > 0) && (first_ix > last_ix)
	  || (step < 0) && (first_ix < last_ix)
	) then (
	  let msg = sprintf "bad array slice [%d..%d] step %d" first_ix last_ix step in
	    raise (EvalArray_error msg)
	) else (
	  (* index relatif du dernier *)
	  let last_rel = abs (last_ix-first_ix) in
	    (* step absolu *)
	  let abs_step = abs step in
	    (* le dernier est-il pris dans la tranche ? *)
	    if((last_rel mod abs_step) <> 0) then
	      warning lxm (
		sprintf
		  "last index out of slice [%d..%d step %d]" first_ix last_ix step)
	    else ();
	    let width = 1 + last_rel/abs_step in
	      (* on force le dernier a tre dans la tranche *)
	    let real_last_ix = first_ix + (width-1) * step in
	      (* (first_ix,last_ix,step,width) *)
	      {
   		se_first = first_ix;
   		se_last = real_last_ix;
   		se_step = step;
   		se_width = width
	      }
	) 
    )
