(** Time-stamp: <modified the 30/01/2008 (at 15:47) by Erwan Jahier> *)


open Lxm
open Errors
open SyntaxTree
open CompiledData


(******************************************************************************)
(** Returns the ident on which the recursion was detected, plus an execution
    stack description. 
*)
exception Recursion_error of (Ident.long as 'id) * (string list as 'stack)

exception BadCheckRef_error

let recursion_error (lxm : Lxm.t) (stack : string list) =
  let rec string_of_stack = function
    | [] -> "" 
    | s::[] -> s
    | s::l  -> s^"\n   > "^(string_of_stack l)
  in
    raise ( Compile_error (lxm,
			   "recursion loop detected:\n   > "^(string_of_stack stack)
   ))


(******************************************************************************)
(* Structure principale *)
type t = {
  src_tab : SyntaxTab.t;
  (* table des defs *)
  types  : (item_key, CompiledData.type_eff  check_flag) Hashtbl.t;
  consts : (item_key, CompiledData.const_eff check_flag) Hashtbl.t;
  nodes  : (node_key, CompiledData.node_eff  check_flag) Hashtbl.t;
  (* table des prov *)
  prov_types  : (item_key, CompiledData.type_eff  check_flag) Hashtbl.t;
  prov_consts : (item_key, CompiledData.const_eff check_flag) Hashtbl.t;
  prov_nodes  : (node_key, CompiledData.node_eff  check_flag) Hashtbl.t
}

(******************************************************************************)
(* exported *)

let (create : SyntaxTab.t -> t) =
fun tbl -> 
  {
    src_tab = tbl;
    types = Hashtbl.create 0;
    consts =  Hashtbl.create 0;
    nodes  = Hashtbl.create 0;
    prov_types = Hashtbl.create 0;
    prov_consts =  Hashtbl.create 0;
    prov_nodes  = Hashtbl.create 0;
  } 

(******************************************************************************)


(******************************************************************************)

(** Type checking + constant checking/evaluation

   This is performed (lazily) by 10 mutually recursive functions:

   types checking
   --------------
   (1) [type_check env type_name lxm]: type check the type id [type_name]
   (3) [type_check_do]: untabulated version of [type_check] (do the real work).

   (2) [type_check_interface]: ditto, but for the interface part
   (2) [type_check_interface_do]: untabulated version (do the real work)

   (4) [solve_type_idref] solves constant reference (w.r.t. short/long ident)

   constants checking
   ------------------
   (5) [const_check env const_name lxm]: eval/check the constant [const_name]
   (7) [const_check_do] : untabulated version (do the real work)

   (6) [const_check_interface]: ditto, but for the interface part
   (6) [const_check_interface_do]: untabulated version (do the real work)

   (8) [solve_const_idref] solves constant reference (w.r.t. short/long ident)

    nb : the 4 functions dealing with constants duplicate 90% of the
    code of the 4 functions dealing with types. It is not easy to
    factorize them out (because of the 10%). I managed to do it for
    [type_check] and [const_check], but even there, it is not that clear
    that it was worthwhile...
*)

(* Before starting, let's define a few utilitary functions. *)

(** Intermediary results are put into a table. This tabulation handling
    is common to type and constant checking, and is performed by the
    2 following functions.

    Since [x] is meant to stand for [type] or [const], those 2
    functions will lead to the definition of 4 functions:
    [type_check], [const_check], [type_check_interface],
    [const_check_interface].
*)
let x_check tab find_x x_check_do lookup_x_eff this x_long_name lxm =
  try lookup_x_eff tab x_long_name lxm 
  with Not_found ->
    Hashtbl.add tab x_long_name Checking;
    let (x_pack,xn) = (Ident.pack_of_long x_long_name, Ident.of_long x_long_name) in
    let x_pack_symbols = SyntaxTab.pack_body_env this.src_tab x_pack in
    let x_def = match find_x x_pack_symbols xn with
      | SymbolTab.Here x_def -> x_def
      | SymbolTab.NotHere _ -> assert false
    in
    let res = x_check_do this x_long_name lxm x_pack_symbols x_pack x_def in
      Hashtbl.replace tab x_long_name (Checked res);
      res

let x_check_interface 
    tab find_x x_check x_check_interface_do lookup_x_eff this x_long_name lxm =
  try lookup_x_eff tab x_long_name lxm
  with Not_found ->
    Hashtbl.add tab x_long_name Checking;
    let (xp,xn) = (Ident.pack_of_long x_long_name, Ident.of_long x_long_name) in
    let xp_prov_symbols_opt = SyntaxTab.pack_prov_env this.src_tab xp in
    let res = (* [xp] migth have no provided symbol table *)
      match xp_prov_symbols_opt with
	| None -> x_check this x_long_name lxm
	    (* if [xp] have no provided symbol table, the whole
	       package is exported. *)
	| Some xp_prov_symbols ->
	    let x_def = match find_x xp_prov_symbols xn with
	      | SymbolTab.Here x -> x
	      | SymbolTab.NotHere _ -> assert false
	    in
	      x_check_interface_do this x_long_name lxm xp_prov_symbols xp x_def
    in
      Hashtbl.replace tab x_long_name (Checked res);
      res

(* Returns the tabulated [type] or [const], if it has already been computed;
   otherwise, raise [Not_found] otherwise. *)
let lookup_x_eff x_label types_tab x_long_name lxm  =
    match Hashtbl.find types_tab x_long_name with
      | Checked res -> res
      | Checking -> raise (Recursion_error (x_long_name, [x_label^(Lxm.details lxm)]))
      | Incorrect -> raise (BadCheckRef_error)

let (lookup_type_eff: (item_key, CompiledData.type_eff check_flag) Hashtbl.t -> 
      Ident.long -> Lxm.t -> CompiledData.type_eff) = 
  lookup_x_eff "type ref "

let (lookup_const_eff:(item_key, CompiledData.const_eff check_flag) Hashtbl.t -> 
      Ident.long -> Lxm.t -> CompiledData.const_eff) = 
  lookup_x_eff "const ref " 


(* And now we can start the big mutually recursive definition... *)

(** Tabulated version of [type_check_do]. *)
let rec (type_check : t -> Ident.long -> Lxm.t -> CompiledData.type_eff) =
  fun this -> 
  x_check this.types SymbolTab.find_type type_check_do lookup_type_eff this

(** Tabulated version of [const_check_do]. *)
and (const_check : t -> Ident.long -> Lxm.t -> CompiledData.const_eff) =
  fun this -> 
  x_check this.consts SymbolTab.find_const const_check_do lookup_const_eff this

(** Tabulated version of [type_check_interface_do]. *)
and (type_check_interface: t -> Ident.long -> Lxm.t -> CompiledData.type_eff) =
  fun this -> 
  x_check_interface 
    this.prov_types SymbolTab.find_type type_check type_check_interface_do 
    lookup_type_eff this

(** Tabulated version of [const_check_interface_do]. *)
and (const_check_interface: t -> Ident.long -> Lxm.t -> CompiledData.const_eff) =
  fun this -> 
    x_check_interface 
      this.prov_consts SymbolTab.find_const const_check const_check_interface_do
      lookup_const_eff this


(* now the real work! *)
and (type_check_interface_do: t -> Ident.long -> Lxm.t -> SymbolTab.t -> 
      Ident.pack_name -> SyntaxTree.type_info srcflaged -> CompiledData.type_eff) =
  fun this type_name lxm prov_symbols pack_name type_def ->
    (* We type check the interface and the body. 
       For non-abstract types, we also check that both effective types are
       the same.  *)
    let body_type_eff = type_check this type_name lxm in
    let prov_type_eff = 
      type_check_do this type_name lxm prov_symbols pack_name type_def
    in
      match prov_type_eff with
	| External_type_eff _ -> prov_type_eff (* no check is necessary *)
	| Bool_type_eff
	| Int_type_eff
	| Real_type_eff
	| Enum_type_eff (_,_)
	| Array_type_eff (_,_)
	| Struct_type_eff (_,_)
	  -> (* Not an abstract type: therefore, we check that [prov_type_eff]
		is the same as [body_type_eff].
	     *)
	    let body_symbols = SyntaxTab.pack_body_env this.src_tab pack_name in
	    let prov_type_eff_body =
	      type_check_do this type_name lxm body_symbols pack_name type_def
	    in
	      assert (prov_type_eff_body = prov_type_eff); (* ??? sont-ce les meme ? *)
	      if prov_type_eff_body = body_type_eff then
		prov_type_eff
	      else
		raise(Compile_error (
			type_def.src,
			("provided type \n\t" ^ 
			   (string_of_type_eff prov_type_eff_body)  ^ 
			   "\n is not compatible with its implementation \n\t" ^ 
			   (string_of_type_eff body_type_eff))))


and (const_check_interface_do: t -> Ident.long -> Lxm.t -> SymbolTab.t -> 
      Ident.pack_name -> SyntaxTree.const_info srcflaged -> CompiledData.const_eff) =
  fun this cn lxm prov_symbols p const_def -> 
    let prov_const_eff = const_check_do this cn lxm prov_symbols p const_def in
    let body_const_eff = const_check this cn lxm in
      match prov_const_eff with
	  (* les consts provided sont toujours abstraites *)
	| Enum_const_eff _ 
	| Extern_const_eff _ -> 
	    (* ... abstrait -> no souci *)
	    let tp = type_of_const_eff prov_const_eff in
	    let tb = type_of_const_eff body_const_eff in
	      if (tp = tb) then body_const_eff 
	      else 
		raise(Compile_error (
			const_def.src,
			("provided constant type \n\t" ^ (string_of_type_eff tp)  ^ 
			   "   is not compatible with its implementation \n\t" ^ 
			   (string_of_type_eff tb) ^ "")
		      ))
 
	| Bool_const_eff _
	| Int_const_eff _
	| Real_const_eff _
	| Struct_const_eff (_,_)
	| Array_const_eff (_,_)
	  ->
	    print_string (
	      (Ident.string_of_long cn) ^ " = " ^
	      (string_of_const_eff prov_const_eff) ^ " is not abstract !");
	    assert false (* pas abstrait => impossible ! *) 

and (type_check_do: t -> Ident.long -> Lxm.t -> SymbolTab.t -> Ident.pack_name ->
      SyntaxTree.type_info srcflaged -> CompiledData.type_eff) =
  fun
    this     (* l'env global*)
    type_name(* juste pour les erreurs de rcursions *)
    lxm      (* ditto *)
    symbols (* table courante des symboles *) 
    p        (* nom du pack courant *)
    type_def 
    -> try (
      (* Solveur d'idref pour les  les appels  eval_type/eval_const *)
      let eval_env = {
	id2const = (solve_const_idref this symbols p);
	id2type = (solve_type_idref this symbols p);
      } 
      in 
      let tinfo = type_def.it in
	match tinfo with
	  | ExternalType s -> External_type_eff (Ident.make_long p s)
	  | AliasedType (s, texp) -> EvalType.f eval_env texp
	  | EnumType (s, clst) -> (
	      let n : Ident.long = Ident.make_long p s in
	      let putpack x = Ident.make_long p x.it in
	      let ncl : Ident.long list = List.map putpack clst in
		Enum_type_eff (n, ncl)
	    )
	  | StructType sti -> (
	      let make_field (fname : Ident.t) = (
		let field_def = Hashtbl.find sti.st_ftable fname in
		let teff = EvalType.f eval_env field_def.it.fd_type in
		  match field_def.it.fd_value with
		    | None -> (fname, teff, None)
		    | Some vexp -> (
			let veff = EvalConst.eval_const eval_env vexp in
			  match veff with
			    | [v] -> (
				let tv = type_of_const_eff v in
				  if (tv = teff) 
				  then (fname, teff, Some v)
				  else 
				    raise 
				      (Compile_error(
					 field_def.src,
					 Printf.sprintf
					   " this field is declared as '%s' but evaluated as '%s'"
					   (string_of_type_eff teff)
					   (string_of_type_eff tv)
				       ))
			      )
			    | [] -> assert false
			    | _::_ -> 
				raise ( Compile_error(
					  field_def.src,
					  "bad field value: tuple not allowed" 
					))
		      ) 
	      ) 
	      in
	      let eff_fields = List.map make_field sti.st_flist in
	      let n = Ident.make_long p sti.st_name in
		Struct_type_eff (n, eff_fields)
	    )
    ) with
	(* capte et complete/stoppe les recursions *)
	Recursion_error (root, stack) -> (
	  if (root = type_name) then recursion_error type_def.src stack
	  else
	    (* on complete la stack *)
	    raise ( Recursion_error (root, ("type ref "^(Lxm.details lxm))::stack))
	)

and (const_check_do : t -> Ident.long -> Lxm.t -> SymbolTab.t -> Ident.pack_name -> 
      SyntaxTree.const_info srcflaged -> CompiledData.const_eff) =
  fun this cn lxm symbols currpack const_def ->
    (* [cn] and [lxm] are used for recursion errors. 
       [symbols] is the current symbol table.
    *)
    try (
      (* Solveur d'idref pour les  les appels  eval_type/eval_const *)
      let eval_env = {
	id2const = (solve_const_idref this symbols currpack) ;
	id2type = (solve_type_idref this symbols currpack) ;
      } 
      in 
      let cinfo = const_def.it in
	match cinfo with
	  | ExternalConst (id, texp) ->
	      Extern_const_eff ((Ident.make_long currpack id), EvalType.f eval_env texp)

	  | EnumConst (id, texp) ->
	      Enum_const_eff ((Ident.make_long currpack id), EvalType.f eval_env texp)

	  | DefinedConst (id, texp_opt, vexp ) -> (
	      match (EvalConst.eval_const eval_env vexp) with
		| [ceff] -> (
		    match texp_opt with
		      | None -> ( ceff)
		      | Some texp -> (
			  let tdecl = EvalType.f eval_env texp in
			  let teff =  type_of_const_eff ceff in
			    if (tdecl = teff ) then ceff
			    else 
			      raise 
				(Compile_error 
				   (const_def.src , 
				    Printf.sprintf
				      " this constant is declared as '%s' but evaluated as '%s'"
				      (string_of_type_eff tdecl)
				      (string_of_type_eff teff)
				   ))
			)
		  )
		| [] -> assert false
		| _::_ -> raise (Compile_error
				   (const_def.src, 
				    "bad constant value: tuple not allowed"))
	    )
    ) with Recursion_error (root, stack) -> (
      (* capte et complete/stoppe les recursions *)
      if (root = cn) then (
	recursion_error const_def.src stack
      ) else (
	(* on complete la stack *)
	raise ( Recursion_error (root, ("const ref "^(Lxm.details lxm))::stack))
      )
    )

(** solving type and constant references *)
and (solve_type_idref : t -> SymbolTab.t -> Ident.pack_name -> Ident.idref -> Lxm.t
      -> CompiledData.type_eff) =
  fun this symbols currpack idr lxm ->
    let s = Ident.name_of_idref idr in
      match Ident.pack_of_idref idr with
	| Some p ->
	    (* rfrence absolue -> on refile le boulot  this *)
	    type_check_interface this (Ident.make_long p s) lxm
	      
	| None -> 
	    (* on passe par symbols pour savoir quoi faire ... *)
	    try 
	      match (SymbolTab.find_type symbols s) with
		| SymbolTab.Here _ ->
		    type_check this (Ident.make_long currpack s) lxm
		      
		| SymbolTab.NotHere fid ->
		    (* rfrence absolue -> on refile le boulot  this *)
		    type_check_interface this fid lxm
		      
	    with Not_found -> 
	      raise(Compile_error(lxm,"unbounded type ident"))
		
and (solve_const_idref : t -> SymbolTab.t -> Ident.pack_name -> Ident.idref ->
      Lxm.t -> const_eff) =
  fun this symbols curpack idr lxm -> 
    let s = Ident.name_of_idref idr in
      match (Ident.pack_of_idref idr) with
	| Some p ->
	    (* rfrence absolue -> on refile le boulot  this *)
	    const_check_interface this (Ident.make_long p s) lxm
	| None ->
	    (* Printf.printf "solve_const_idref %s\n" s ; *)
	    (* on passe par symbols pour savoir quoi faire ... *)
	    try
	      match (SymbolTab.find_const symbols s) with
		| SymbolTab.Here _ ->
		    const_check this (Ident.make_long curpack s) lxm
		| SymbolTab.NotHere fid ->
		    (* rfrence absolue -> on refile le boulot  this *)
		    const_check_interface this fid lxm
	    with Not_found ->
	      raise(Compile_error(lxm,"unbounded const ident"))
	 
  

(******************************************************************************)
(* exported *)
let (do_node: t -> CompiledData.node_key -> CompiledData.node_eff) =
  fun this nk ->
(* XXX finish me ! *)
    let itl = [] in
    let otl = [] in
    let icl = [] in
    let ocl = [] in
      
    assert false;
      {
	nf_key = nk ;
	nf_in_types = itl;
	nf_out_types = otl;
	nf_in_formal_clocks = icl;
	nf_out_formal_clocks = ocl;
      }

(*-------------------------------------------------------------------------
 Test/debug
---------------------------------------------------------------------------*)


let test (this: t) = (
  (* src_tab : SyntaxTab.t; *)
  let testpack pack_name = (
    Verbose.printf " * package %s\n" (Ident.pack_name_to_string pack_name);
    let prov_symbols = 
      match SyntaxTab.pack_prov_env this.src_tab pack_name with
	| Some tab -> tab
	| None -> SyntaxTab.pack_body_env this.src_tab pack_name
    in
    let test_item label x_check_interface string_of_x_eff item item_def =
      match item_def with
	| SymbolTab.Here _item_def ->
	    let x_eff = 
	      x_check_interface this (Ident.make_long pack_name item) Lxm.dummy  
	    in
	      Printf.printf "\t\t%s %s = %s\n" 
		label (Ident.to_string item) (string_of_x_eff x_eff)
	  
	| SymbolTab.NotHere item_def ->
	    Printf.printf "\t\t%s %s = %s (imported)\n" 
	      label (Ident.to_string item) (Ident.string_of_long item_def);
    in
    let test_types = test_item "type" type_check_interface string_of_type_eff in
    let test_constants = test_item "const" const_check_interface string_of_const_eff in
(*     let test_operators = test_item "" in *)
      Verbose.print_string "\tExported types:\n";
      SymbolTab.iter_types prov_symbols test_types ;
      Verbose.print_string "\tExported constants:\n";
      SymbolTab.iter_consts prov_symbols test_constants;
(*       Verbose.print_string "\tExported operators:\n"; *)
(*       SymbolTab.iter_consts prov_symbols test_operators *)
  )
  in
  let plist = SyntaxTab.pack_list this.src_tab in
    Verbose.print_string "*** Dump the exported items of the packages.\n";
    List.iter testpack plist
)
