(** Time-stamp: <modified the 30/01/2008 (at 16:20) by Erwan Jahier> *)

(** Here follows a description of the different modules used by this lus2lic compiler.

(1) First of all, the Lustre files are parsed,

   lexer.mll  
   parser.mly
   parserUtils.ml
   lxm.mli/ml

which results into a parse tree containing raw source expressions.

   syntaxTree.ml  -> should rather be called rawSyntaxTab.ml ?
   syntaxTreedump.ml

(2) Then, we perform reference checking at module level + model expansion.

   syntaxTab.mli/ml
      syntaxTabUtil.ml/mli
      expandPack.mli/ml
      symbolTab.mli/ml (type/const/oper)

syntaxTab is a kind of layer above syntaxTree to make things easier afterwards.


(3) Finally, the compilation (type checking+const/type evaluation) is performed.

   compile.ml
      lazyCompiler.mli/ml
        evalConst.mli/ml
        evalType.mli/ml
   compiledData.ml
    

Some misc (eponymous) modules are used along the way.
   errors.ml 
   verbose.mli/ml
   version.ml
   ident.ml
*)


open Verbose
open SyntaxTree
open Lxm
open Errors
open Parsing
open Format
open Global

let usage_msg =  
  "usage: "^(Version.tool)^" [options] <file> | "^(Version.tool)^" -help"

(*---------------------------------------------------------
Les args sont des variables GLOBALES
---------------------------------------------------------*)

let print_version = function (x: unit) -> (
  print_string (Version.str ^ "\n")
)

let rec set_infile = function (x:string) -> (
  if ( _lus2lic_ARGS.infile = "" ) then (
    _lus2lic_ARGS.infile <- x
  ) else (
    (* ??? XXX why ??? *)
    Arg.usage
      arg_list 
      ("unexpected argument \""^x^"\"\n"^usage_msg) ;
    exit 1
  )
)
and arg_list = [
  ( "-version", Arg.Unit(function x -> print_version () ; exit 0),
    "  print the current version then exit"
  );
  ( "-n", Arg.String(function x -> _lus2lic_ARGS.main_node <- x ; ()),
    "  <node> specify the main node"
  );
  ( "-tlex", Arg.Unit(function x -> _lus2lic_ARGS.runmode <- "tlex" ; ()),
    "  test lexical analysis"
  );
  ( "-tcheck", Arg.Unit(function x -> _lus2lic_ARGS.runmode <- "tcheck" ; ()),
    "  test static analysis"
  );
  ( "-dump", Arg.Unit(function x -> _lus2lic_ARGS.runmode <- "dump" ; ()),
    "  parse and dump the internal data"
  );
  ( "-v", Arg.Unit (function vl -> Verbose.set_level 1 ),
    "  set verbose mode on (i.e., verbose level = 1)"
  );
  ( "-vl", Arg.Int(function vl -> Verbose.set_level vl ),
    "  <int> set verbose level"
  )
]
and
    parse_args () = (
      Arg.parse arg_list (* liste des options *)
	set_infile       (* arg par defaut = fichier d'entree *)
	usage_msg        (* message d'erreur *)
      ;
      ()
    )

let test_lex ( lexbuf ) = (
  let tk = ref (Lexer.lexer lexbuf) in (
      while !tk <> Parser.TK_EOF do
	match (Lexer.token_code !tk) with 
	    ( co , lxm ) ->
	      printf "%s : %15s = \"%s\"\n"
		(Lxm.position lxm) co (Lxm.str lxm) ;
	      tk := (Lexer.lexer lexbuf)
      done
    )	
)

(* Retourne un parse_tree *)
let lus_load lexbuf = (
  (Parser.sxLusFile Lexer.lexer lexbuf)
) 

(* Dump d'un packbody *)
let dump_body (pkg: SyntaxTree.packbody) = (
  let os = Format.formatter_of_out_channel stdout in
    SyntaxTreeDump.packbody os pkg	
)

(* Dump d'un name-space, pack ou modele ...  *)
let dump_ns (ns: SyntaxTree.pack_or_model) = (
  let os = Format.formatter_of_out_channel stdout in
    match ns with
	NSPack pf -> (
	  (* Verbose.printf (lazy ("DUMP PACKDEF\n")); *)
	  SyntaxTreeDump.packinfo os pf
	) 
      | NSModel mf -> (
	  (* Verbose.printf (lazy ("DUMP MODDEF\n")); *)
	  SyntaxTreeDump.modelinfo os mf
	) 
)

  (*
    Lance le parser et renvoie la liste name-spaces d'entre.	
    Dans le cas d'un fichier sans package, on lui donne
    comme nom le basename de infile.
  *)
let get_source_list infile lexbuf = (
  match (lus_load lexbuf) with
      PRPackBody pbdy -> 
	(
	  let nme = Filename.chop_extension (Filename.basename infile) in
	  let pi = 
	    SyntaxTree.give_pack_this_name (Ident.pack_name_of_string nme) pbdy 
	  in
	    [ NSPack (Lxm.flagit pi Lxm.dummy) ]
	) 
    | PRPack_or_models nsl -> nsl
)

let main = (
  (* Compile.init_appli () ; *)
  parse_args () ;
  if (_lus2lic_ARGS.infile = "") then (
    Arg.usage arg_list usage_msg ;
    exit 1
  ) ;
  try (
    let inchannel = open_in _lus2lic_ARGS.infile in
    let lexbuf = Lexing.from_channel inchannel in
      match _lus2lic_ARGS.runmode with
	| "tlex"   -> test_lex lexbuf
	| "dump" ->
	    let nsl = get_source_list _lus2lic_ARGS.infile lexbuf in
	      List.iter dump_ns nsl ;
	      
	|  _ -> (
	     (* Compile.compile (pkg_load lexbuf) _lus2lic_ARGS.main_node  *)
	     let nsl = get_source_list _lus2lic_ARGS.infile lexbuf in
	     let nodidf = Ident.idref_of_string _lus2lic_ARGS.main_node in
	       Compile.doit nsl nodidf 
	   )
  ) with
      Sys_error(s) ->
	prerr_string (s^"\n") ;
	exit 1 
    | Global_error s ->
       print_global_error s ;
	  exit 1
    | Parse_error ->
	print_compile_error (Lxm.last_made ()) "syntax error";
	exit 1
    | Compile_error(lxm,msg) -> 
	print_compile_error lxm msg ;
	exit 1
    | Assert_failure (file, line, col)  -> 
	prerr_string (
	  "\n*** oops: an internal error occurred in file "^ file ^ ", line " ^ 
	    (string_of_int line) ^ ", column " ^
	    (string_of_int col) ^ "\n*** when compiling lustre program " ^
	    _lus2lic_ARGS.infile ^ "\n") ;
	exit 2

	(* | Compile_node_error(nkey,lxm,msg) -> ( *)
	(* print_compile_node_error nkey lxm msg ; *)
	(* exit 1 *)
	(* ) *)
	(* | Global_node_error(nkey,msg) -> ( *)
	(* print_global_node_error nkey msg ; *)
	(* exit 1 *)
	(* ) *)

)
